/*
  This implements the user space software for programming the Cygnal C8051F320/1
  via the C2 interface.
  Written by Wojciech M. Zabolotny (wzab@ise.pw.edu.pl)
  Published under the GPL (Gnu Public License) version 2.0 or later
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "c2_ioctls.h"

//Status bits (as read from Address register)
#define ST_IN_BUSY ( 2 )
#define ST_OUT_READY ( 1 )

//Table 3. C2 FLASH Programming Commands
#define  FLASH_BLOCK_WRITE      ( 0x07 )
#define  FLASH_BLOCK_READ       ( 0x06 )
#define  FLASH_PAGE_ERASE       ( 0x08 )
#define  DEVICE_ERASE           ( 0x03 )

// C2 addresses
#define DEVICE_ID (0x00 )
#define FPDAT ( 0xb4 )
#define FPCTL ( 0x02)

#define min(a,b) ( ((a)<(b)) ? (a) : (b) )
int progfile;
unsigned char buf[100];

//Delay
void udelay(int usecs)
{
  //At the moment we use usleep, however it is obsolete.
  //Then we should switch to nanosleep.
  usleep(usecs);
}

/* C2 utility functions */
int c2_device_id()
{
  unsigned char c;  unsigned long l;
  c=DEVICE_ID;
  ioctl(progfile,C2_IOCAWRITE,&c);
  ioctl(progfile,C2_IOCD1READ,&l);
  return l;  
}

void c2_wait_in_busy()
{
  unsigned char c;
  while( ioctl(progfile,C2_IOCAREAD,&c), c & ST_IN_BUSY) {};
}

void c2_wait_out_ready()
{
  unsigned char c;
  while (ioctl(progfile,C2_IOCAREAD,&c), !(c & ST_OUT_READY)) {};
}

int c2_block_write(unsigned int address, unsigned char * data, int len)
{
  unsigned char c;
  unsigned long l;
  int i;
  //An127, page 5!!!
  c=FPDAT;
  ioctl(progfile,C2_IOCAWRITE,&c);
  c=FLASH_BLOCK_WRITE;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();  //Wait until ST_InBusy=0
  c2_wait_out_ready(); //Wait until ST_OutReady=1
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status1 0x%2.2lx instead of 0x0d in c2_block_write\n",l);
    return -1;
  }
  c=address >> 8;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c=address & 0xff;  //Write address low byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  //Write length code to FPDAT
  if(len==256)
    c=0;
  else
    c=len;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c2_wait_out_ready();
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status2 0x%2.2lx instead of 0x0d in c2_block_write\n",l);
    return -1;
  }
  for(i=0;i<len;i++) {
    ioctl(progfile,C2_IOCD1WRITE,&data[i]);
    c2_wait_in_busy();
  }
  c2_wait_out_ready();
  return 0;
}

int c2_block_read(int address, unsigned char * data, int len)
{
  //An127, page 6!!!
  unsigned char c;
  unsigned long l;
  int i;
  c=FPDAT;
  ioctl(progfile,C2_IOCAWRITE,&c);
  c=FLASH_BLOCK_READ;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();  //Wait until ST_InBusy=0
  c2_wait_out_ready(); //Wait until ST_OutReady=1
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status1 0x%2.2lx instead of 0x0d in c2_block_read\n",l);
    return -1;
  }
  c=address >> 8;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c=address & 0xff;  //Write address low byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  //Write length code to FPDAT
  if(len==256)
    c=0;
  else
    c=len;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c2_wait_out_ready();
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status2 0x%2.2lx instead of 0x0d in c2_block_read\n",l);
    return -1;
  }
  for(i=0;i<len;i++) {
    c2_wait_out_ready();
    ioctl(progfile,C2_IOCD1READ,&l);
    data[i]=l;
    //printf("read: %x\n",l);
  }
  return 0;
}

int c2_page_erase(int page)
{
  unsigned char c;
  unsigned long l;
  int i;
  //An127, page 7!!!
  c=FPDAT;
  ioctl(progfile,C2_IOCAWRITE,&c);
  c=FLASH_PAGE_ERASE;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();  //Wait until ST_InBusy=0
  c2_wait_out_ready(); //Wait until ST_OutReady=1
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status1 0x%2.2lx instead of 0x0d in c2_page_erase\n",l);
    return -1;
  }
  c=page;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c2_wait_out_ready(); //Wait until ST_OutReady=1
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status2 0x%2.2lx instead of 0x0d in c2_page_erase\n",l);
    return -1;
  }
  c=0x00;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c2_wait_out_ready();
  return 0;
}

int c2_device_erase()
{
  //An127, page 9!!!
  unsigned char c;
  unsigned long l;
  int i;
  c=FPDAT;
  ioctl(progfile,C2_IOCAWRITE,&c);
  c=DEVICE_ERASE;
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();  //Wait until ST_InBusy=0
  c2_wait_out_ready(); //Wait until ST_OutReady=1
  ioctl(progfile,C2_IOCD1READ,&l); //Read the status and check if it is 0xd
  if(l != 0xd) {
    fprintf(stderr,"Status1 0x%2.2x instead of 0x0d in c2_block_read\n",l);
    return -1;
  }
  c=0xde;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c=0xad;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c=0xa5;  //Write address high byte to FPDAT
  ioctl(progfile,C2_IOCD1WRITE,&c);
  c2_wait_in_busy();
  c2_wait_out_ready();
  return 0;
}

void c2_init()
{
  unsigned char c;
  unsigned long l;
  ioctl(progfile,C2_IOCRESET,NULL);
  udelay(2);
  c=FPCTL;
  ioctl(progfile,C2_IOCAWRITE,&c);
  l=2;
  ioctl(progfile,C2_IOCD1WRITE,&l);
  l=1;
  ioctl(progfile,C2_IOCD1WRITE,&l);
  udelay(20000);
}


// Function for reading the file into memory buffer
// we can read all the memory data at once, no need
// for sofisticated buffering the memory in FLASH uC is so small ;-)
// This function allocates a buffer, so we are responsible to free it
// after use
// Parameters: filename - obvious meaning
//   buf - address of the pointer to the data buffer (we fill it after allocation)
int readfile(char * filename, unsigned char * * buf)
{
  int result=0;
  int dfile=-1, dsize=0;
  *buf = NULL;
  dfile=open(filename,O_RDONLY);
  if(dfile<0) {
    perror(filename);
    return -1;
  }
  dsize=lseek(dfile,0,SEEK_END);
  if(dsize<0) {
    perror("lseek");
    result=-2;
    goto cleanup;
  }
  *buf = malloc(dsize);
  if(*buf == NULL) {
    perror("data buffer malloc");
    result=-3;
    goto cleanup;
  }
  if(lseek(dfile,0,SEEK_SET) != 0) {
    perror("lseek");
    result=-4;
    goto cleanup;
  }
  if(read(dfile,*buf,dsize)!=dsize) {
    fprintf(stderr,"I can't read data\n");
    result=-5;
    goto cleanup;
  }
  return dsize;
 cleanup:
  if(dfile) close(dfile);
  if(*buf) {
    free(*buf);
    *buf = NULL;
  }
  return result;
}

// Function for writing the contents of a memory buffer into the file
int writefile(char * filename, int dsize, unsigned char * buf)
{
  int result=0;
  int dfile=0;
  int wpos=0;
  dfile=open(filename,O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
  if(dfile<0) {
    perror(filename);
    return -1;
  }
  while(wpos<dsize) {
    int nwritten=write(dfile,buf+wpos,dsize-wpos);
    if(nwritten<0) {
      perror("write");
      fprintf(stderr,"I can't write data\n");
      result=-5;
      goto cleanup;
    }
    wpos+=nwritten;
  }
  return 0;
 cleanup:
  if(dfile>0) close(dfile);
  return result;
}

//Function for reading the device's FLASH memory
//this function allocates a data buffer, so we need to free it after use
int readprg(int psize, unsigned char * * buf)
{
  * buf = NULL;
  int result=0;
  int i;
  //First - allocate  the buffer for the data
  (* buf)=malloc(psize);
  if((* buf)==NULL) {
    fprintf(stderr,"I can't allocate the memory\n");
    result = -1;
    goto cleanup;
  }
  for(i=0;i<psize;) {
    int nread = min(256,psize-i); //number of bytes to read
    if(c2_block_read(i, (*buf)+i, nread)) {
      fprintf(stderr,"c2_block_read error in readprg\n");
      result = -1;
      goto cleanup;
    }
    i+=nread;
  } 
  return 0;
 cleanup:
  if(*buf) {
    free(*buf);
    *buf = NULL;
  }
  return result;
}

// Function verifies the device's memory contents with the data buffer
// returns 0 on success.
int vrfyprg(int dsize, unsigned char * fbuf)
{
  unsigned char * buf = NULL;
  int i=0,result=0,cmp_err=1;
  //We reed the data from the device
  if(readprg(dsize,&buf)) {
    fprintf(stderr,"I was not able to allocate memory in the "
	    "\"vrfyprg\" function\n");
    result=-1;
    goto cleanup;
  }
  //The data have been read, now we can verify it
  for(i=0;i<dsize;i++) {
    if(fbuf[i] != buf[i]) {
      fprintf(stderr,
	      "Comparison error at 0x%4.4x: device:0x%2.2x file:0x%2.2x \n",
	      i,buf[i],fbuf[i]);
      result=1;
    }
  }
 cleanup:
  if(buf) free(buf);
  return result;
}

//Function for printing the buffer
void printbuf(int dsize, unsigned char * fbuf)
{
  int i;
  for(i=0;i<dsize;i++){
    if(i%8==0) {
      printf("\n%2.2x : ",i);
    }
    printf("%2.2x ",(int) fbuf[i]);
  }
  printf("\n");
}

//Function for writing the program (read from the file) to the device's memory
int writeprg(char * filename)
{
  int dsize,i;
  unsigned char * fbuf=NULL;
  unsigned char * dbuf=NULL;
  int result=0;
  unsigned char needs_erase=0;
  //First read the data to be programmed
  dsize=readfile(filename,&fbuf);
  if(dsize<=0) {
    fprintf(stderr,"I can't read from %s the data to be programmed\n",
	    filename);
    goto cleanup;
  }
  //Then read the data already located in the device
  if(readprg(dsize,&dbuf)) {
    fprintf(stderr,"I can't read data from the device\n");
    goto cleanup;
  }
  //Now we check whether the device needs erasing
  //(If any bit is set to 1 in the file data and to 0 in the device,
  //we need to erase the device, however later a page based approach should
  //be used instead)
  for(i=0;i<dsize;i++) {
    if(fbuf[i] & ~dbuf[i]) needs_erase=1;
  }
  //Erase the device if it is necessary
  //Again - later we should check and erase page by page
  if(needs_erase) {
    printf("device needs erasing!\n");
    c2_device_erase();
    printf("erased\n");
  }
  //And now let's program a new data
  //Well later we can change it to program only the values which require modification.
  //But this is only a first, draft version!!!
  for(i=0;i<dsize;) {
    int nprog = min(256,dsize-i); //number of bytes to program
    if(c2_block_write(i, &(fbuf[i]), nprog)) {
      fprintf(stderr,"c2_block_read error in writeprg\n");
      result = -1;
      goto cleanup;
    }
    i+=nprog;
  }
  //finally we should verify the data
  free(dbuf); dbuf=NULL;
  result=vrfyprg(dsize,fbuf);
  if(!result) {
    printf("Succesfully verified %d bytes\n",dsize);
  }
 cleanup:
  if(dbuf) free(dbuf);
  if(fbuf) free(fbuf);
  return result;
}

//Function for checking the program with the file
int checkprg(char * filename)
{
  int dsize,i;
  unsigned char * fbuf=NULL;
  int result=0;
  unsigned char needs_erase=0;
  //First read the data to be programmed
  dsize=readfile(filename,&fbuf);
  if(dsize<=0) {
    fprintf(stderr,"I can't read from %s the data to be programmed\n",
	    filename);
    goto cleanup;
  }
  result=vrfyprg(dsize,fbuf);
  if(!result) {
    printf("Succesfully verified %d bytes\n",dsize);
  }
 cleanup:
  if(fbuf) free(fbuf);
  return result;
}

//Function for storing the device's memory contents into the file
int storeprg(char * filename)
{
  int result=0;
  unsigned char * buf = NULL;
  int psize = 0x3e00; //Locations between 3e00 end 3fff are reserved!!!
  if(readprg(psize,&buf)) {
    fprintf(stderr,"I can't read the data from the device\n",
	    filename);
    goto cleanup;    
  }
  //Now we've read data from the device, so we can store them into file
  result=writefile(filename,psize,buf);
 cleanup:
  if(buf) free(buf);
  return result;
}

void disp_usage()
{
  printf("The correct calling syntax is:\n"
	 "cygprog command file\n"
	 "The command can be:\n"
	 "\tp - programm\n"
	 "\tv - verify\n"
	 "\tr - read and store to the file\n");
}

int main(int argc, char * argv[])
{
  //In the first version we service only the binary files
  //starting from "0" address
  //You can convert the "ihx" files produced by the sdcc
  //into the binary files using the "srec_cat" tool
  //eg.: srec_cat -Output test.bin -Binary test.ihx -Intel
  unsigned char c=0;
  long lb;
  int ad=0;
  int i;
  int result=1;
  if(argc==3) {
   progfile=open("./cygnal",O_RDWR);
   if(progfile<0) {
     perror("I can't open device file");
     exit(1);
   }
   //We have opened the device file
   printf("The device id: %x\n",c2_device_id());
   c2_init();
   printf("I've reset the device\n");
   sleep(1);
   //Now we need to check the user command
   if(!strcmp("p",argv[1])) {
     //OK so we are programming
     result=writeprg(argv[2]);
   } else if(!strcmp("v",argv[1])) {
     //OK so we are verifying
     result=checkprg(argv[2]);
   } else if(!strcmp("r",argv[1])) {
     //We are reading the program from the device
     result=storeprg(argv[2]);
   } else {
     //Display the info about available commands
     disp_usage();
   }
 } else {
   disp_usage();
 }
 if(progfile) close(progfile);
 return result;
}
